--
-- Main module, provides multiple main assists handling methods
--


local Roster = AceLibrary("Roster-2.1")
local L = AceLibrary("AceLocale-2.2"):new("BiGOpt")


-- Constants

local BIGOPT_MACRO_NAME = "BiGOptAssistMacro"
local BIGOPT_COMM_PREFIX = "BiGOpt"

-- Initialization methods

BiGOpt = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceComm-2.0", "AceDebug-2.0", "FuBarPlugin-2.0")

BiGOpt:SetCommPrefix(BIGOPT_COMM_PREFIX)
BiGOpt.OnMenuRequest = {
    type = "group",
    args = {
        add = {
            type = "text",
            name = L["Add main assist"],
            desc = L["add a player to the main assists list"],
            usage = L["/bo add Player"],
            get = false,
            set = "addMainAssist",
            order = 10
        },
        list = {
            type = "execute",
            name = L["List main assists"],
            desc = L["print the main assists list"],
            func = "printMainAssists",
            order = 20
        },
        clear = {
            type = "execute",
            name = L["Clear main assists"],
            desc = L["clear the main assists list"],
            func = "clearMainAssists",
            order = 30
        },
        current = {
            type = "execute",
            name = L["Print current main assist"],
            desc = L["print the current main assist"],
            func = "printCurrentMainAssist",
            order = 40
        },
        query = {
            type = "execute",
            name = L["Query main assists from group"],
            desc = L["manually query the main assists list"],
            func = "query",
            order = 50
        },
        savesg = {
            type = "execute",
            name = L["Save subgroups"],
            desc = L["save subgroups composition"],
            func = "saveSubgroups",
            order = 100
        },
        restoresg = {
            type = "execute",
            name = L["Restore subgroups"],
            desc = L["restore the subgroups composition"],
            func = "restoreSubgroups",
            order = 110
        },
        clearsg = {
            type = "execute",
            name = L["Clear subgroups"],
            desc = L["clear the saved subgroups"],
            func = "clearSubgroups",
            order = 120 
        },
        sg = {
            type = "execute",
            name = L["Print saved subgroup"],
            desc = L["display saved subgroups composition"],
            func = "printSubgroups",
            order = 130
        },
        debug = {
            type = "toggle",
            name = L["Debug"],
            desc = L["toggles showing debug information"],
            get = "getDebug",
            set = "setDebug",
            order = 1000
        },
        
    }
}

function BiGOpt:OnInitialize()
    self.debug = false
    self.currentMainAssist = nil
    self.mainAssists = {}
    self.nbMainAssists = 0
    self.subgroups = BiGOptSubGroups:new()
    self.subgroups:setConsoleRenderer(self)
    self.pendingMainAssist = nil

    -- Register slash commands
    self:RegisterChatCommand({"/bigopt", "/bo"}, BiGOpt.OnMenuRequest)

    -- Initialiaze AceComm
    self:RegisterComm(BIGOPT_COMM_PREFIX, "RAID", "receiveMessage")
    self:RegisterComm(BIGOPT_COMM_PREFIX, "WHISPER", "receiveMessage")

    -- Register events
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:RegisterEvent("PLAYER_LEAVE_COMBAT")
    self:RegisterEvent("RosterLib_RosterUpdated")
    self:ScheduleRepeatingEvent(self.update, 1, self)
end

function BiGOpt:OnEnable()
end

function BiGOpt:OnDisable()
end

-- Public interface

function BiGOpt:clearMainAssists(silent)
    silent = silent or false
    if not silent and not self:canModifyList() then 
        self:Print(L["you must be promoted to do that"])
        return
    end

    self.mainAssists = {}
    self.nbMainAssists = 0
    if not silent then
        self:Print(L["main assists list cleared"])
        self:SendCommMessage("RAID", "clear")
    end
end

function BiGOpt:addMainAssist(name, silent)
    name = self:formatName(name)
    silent = silent or false
    if not silent and not self:canModifyList() then 
        self:Print(L["you must be promoted to do that"])
        return
    end

    for _, player in pairs(self.mainAssists) do
        if player == name then
            self:Print(string.format(L["%s is already a main assist"], name))
            return
        end
    end
    self.nbMainAssists = self.nbMainAssists + 1
    self.mainAssists[self.nbMainAssists] = name
    if not silent then 
        self:Print(string.format(L["added %s to the main assists list"], name))
        self:SendCommMessage("RAID", "add", name)
    end
end

function BiGOpt:printMainAssists()
    if self.nbMainAssists ~= 0 then
        self:Print(L["main assists list:"])
        for _, name in pairs(self.mainAssists) do
            self:Print("   - " .. name)
        end
    else
        self:Print(L["main assists list is empty"])
    end
end

function BiGOpt:setDebug(value)
    self.debug = value
    self:SetDebugging(self.debug)
end

function BiGOpt:getDebug()
    return self.debug
end

function BiGOpt:query()
    self:SendCommMessage("RAID", "query_list")
end

function BiGOpt:printCurrentMainAssist()
    if self.currentMainAssist ~= nil then 
        self:Print(string.format(L["current main assist is %s"], self.currentMainAssist))
    else
        self:Print(string.format(L["no main assist"], self.currentMainAssist))
    end
end

function BiGOpt:printSubgroups()
    self.subgroups:renderSaved(self)
end

function BiGOpt:saveSubgroups()
    self.subgroups:save()
    self:Print(L["saved subgroups composition"])
end

function BiGOpt:restoreSubgroups()
    self.subgroups:restore()
    self:Print(L["restored subgroups composition"])
end

function BiGOpt:clearSubgroups()
    self.subgroups:clearSavedSubgroups()
    self:Print(L["cleared saved subgroups"])
end

-- Protected methods

function BiGOpt:update()
    local unit, distance, minDistance, nearestMainAssist

    if GetNumRaidMembers == 0 then 
        return 
    end

    minDistance = 1000000
    for _, name in pairs(self.mainAssists) do
        unit = Roster:GetUnitIDFromName(name)
        if unit ~= nil then
            if self:isAssistable(unit) then
                if CheckInteractDistance(unit, 4) then
                    distance = 30
                else
                    distance = 10000  -- just an arbitrary big distance
                end
                if distance < minDistance then
                    minDistance = distance
                    nearestMainAssist = name
                end
            end
        end
    end

    if nearestMainAssist ~= nil then
        if nearestMainAssist ~= self.currentMainAssist then 
            if not InCombatLockdown() then
                self:updateMainAssist(nearestMainAssist)
            else
                self.pendingMainAssist = nearestMainAssist
            end
        end
    end
end

function BiGOpt:updateMainAssist(name)
    EditMacro(self.macroIndex, BIGOPT_MACRO_NAME, 1, "/assist " .. name, 1)
    self:Debug(L["new main assist is %s"], name)
    self.currentMainAssist = name
end

function BiGOpt:isAssistable(unit)
    return self:isInRange(unit) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit)
end

function BiGOpt:isInRange(unit)
    if UnitIsVisible(unit) then
        return true
    end
    return false
end

function BiGOpt:formatName(name)
    return string.upper(string.sub(name, 1, 1)) .. string.lower(string.sub(name, 2))
end

function BiGOpt:receiveMessage(prefix, sender, distribution, action, arg1, arg2)
    self:Debug(L["got message %s(%s, %s)"], action, arg1, arg2)

    if action == "add" then
        self:addMainAssist(arg1, true)
    elseif action == "clear" then
        self:clearMainAssists(true)
    elseif action == "query_list" then
        self:SendCommMessage("WHISPER", sender, "give_list", self.mainAssists, self.nbMainAssists)
    elseif action == "give_list" then
        self:setMainAssists(arg1, arg2)
    end
end

function BiGOpt:canModifyList()
    return IsRaidLeader() or IsRaidOfficer()
end

function BiGOpt:setMainAssists(mainAssists, nbMainAssists)
    self.mainAssists = mainAssists
    self.nbMainAssists = nbMainAssists
end

function BiGOpt:cleanup()
    -- Cleanup the main assists list, removing players that are not in the party
    local newMainAssists = {}
    local newNbMainAssists = 0

    for key, name in pairs(self.mainAssists) do
        unit = Roster:GetUnitIDFromName(name)
        if unit ~= nil then
            newMainAssists[key] = name
            newNbMainAssists = newNbMainAssists + 1
        end
    end

    self.mainAssists = newMainAssists
    self.nbMainAssists = newNbMainAssists
end

-- Event handlers

function BiGOpt:PLAYER_ENTERING_WORLD()
    -- Search for the BiGOpt macro or create it
    self.macroIndex = GetMacroIndexByName(BIGOPT_MACRO_NAME)
    if self.macroIndex == 0 then
        if name ~= BIGOPT_MACRO_NAME then
            self.macroIndex = CreateMacro(BIGOPT_MACRO_NAME, 1, "", 1, 0)
        end
    end

    -- If player is already in a raid query the main assists list
    if UnitInRaid("player") then
        self:query()
    end
end

function BiGOpt:PLAYER_LEAVE_COMBAT()
    if self.pendingMainAssist ~= nil then
        self:updateMainAssist(self.pendingMainAssist)
        self.pendingMainAssist = nil
    end
end

function BiGOpt:RosterLib_RosterUpdated()
    self.subgroups:restoreLeadership()
    if self.subgroups:complete() then
        self.subgroups:restore()
    end
end
